@@ -2,19 +2,21 @@
 // for details. All rights reserved. Use of this source code is governed by a
 // BSD-style license that can be found in the LICENSE file.
 
-// All imports must be in all FFI patch files to not depend on the order
-// the patches are applied.
 import "dart:_internal" show patch;
 import 'dart:typed_data';
 import 'dart:isolate';
+import 'dart:typed_data';
 
-DynamicLibrary _open(String path) native "Ffi_dl_open";
-DynamicLibrary _processLibrary() native "Ffi_dl_processLibrary";
-DynamicLibrary _executableLibrary() native "Ffi_dl_executableLibrary";
+@pragma("vm:external-name", "Ffi_dl_open")
+external DynamicLibrary _open(String path);
+@pragma("vm:external-name", "Ffi_dl_processLibrary")
+external DynamicLibrary _processLibrary();
+@pragma("vm:external-name", "Ffi_dl_executableLibrary")
+external DynamicLibrary _executableLibrary();
 
 @patch
 @pragma("vm:entry-point")
-class DynamicLibrary {
+final class DynamicLibrary {
   @patch
   factory DynamicLibrary.open(String path) {
     return _open(path);
@@ -27,12 +29,19 @@ class DynamicLibrary {
   factory DynamicLibrary.executable() => _executableLibrary();
 
   @patch
-  Pointer<T> lookup<T extends NativeType>(String symbolName)
-      native "Ffi_dl_lookup";
+  @pragma("vm:external-name", "Ffi_dl_lookup")
+  external Pointer<T> lookup<T extends NativeType>(String symbolName);
 
-  // TODO(dacoharkes): Expose this to users, or extend Pointer?
-  // https://github.com/dart-lang/sdk/issues/35881
-  int getHandle() native "Ffi_dl_getHandle";
+  @patch
+  @pragma("vm:external-name", "Ffi_dl_providesSymbol")
+  external bool providesSymbol(String symbolName);
+
+  @pragma("vm:external-name", "Ffi_dl_getHandle")
+  external int getHandle();
+
+  @patch
+  @pragma("vm:external-name", "Ffi_dl_close")
+  external void close();
 
   @patch
   bool operator ==(Object other) {
@@ -50,9 +59,10 @@ class DynamicLibrary {
   Pointer<Void> get handle => Pointer.fromAddress(getHandle());
 }
 
+@patch
 extension DynamicLibraryExtension on DynamicLibrary {
   @patch
-  DS lookupFunction<NS extends Function, DS extends Function>(
-          String symbolName) =>
+  DS lookupFunction<NS extends Function, DS extends Function>(String symbolName,
+          {bool isLeaf = false}) =>
       throw UnsupportedError("The body is inlined in the frontend.");
 }
