@@ -8,22 +8,26 @@ part of dart.developer;
 ///
 /// If the RPC was successful, use [ServiceExtensionResponse.result], otherwise
 /// use [ServiceExtensionResponse.error].
-class ServiceExtensionResponse {
-  final String _result;
-  final int _errorCode;
-  final String _errorDetail;
+final class ServiceExtensionResponse {
+  /// The result of a successful service protocol extension RPC.
+  final String? result;
+
+  /// The error code associated with a failed service protocol extension RPC.
+  final int? errorCode;
+
+  /// The details of a failed service protocol extension RPC.
+  final String? errorDetail;
 
   /// Creates a successful response to a service protocol extension RPC.
   ///
   /// Requires [result] to be a JSON object encoded as a string. When forming
   /// the JSON-RPC message [result] will be inlined directly.
   ServiceExtensionResponse.result(String result)
-      : _result = result,
-        _errorCode = null,
-        _errorDetail = null {
-    if (_result is! String) {
-      throw new ArgumentError.value(_result, "result", "Must be a String");
-    }
+      : result = result,
+        errorCode = null,
+        errorDetail = null {
+    // TODO: When NNBD is complete, delete the following line.
+    checkNotNullable(result, "result");
   }
 
   /// Creates an error response to a service protocol extension RPC.
@@ -33,32 +37,14 @@ class ServiceExtensionResponse {
   /// encoded as a string. When forming the JSON-RPC message [errorDetail] will
   /// be inlined directly.
   ServiceExtensionResponse.error(int errorCode, String errorDetail)
-      : _result = null,
-        _errorCode = errorCode,
-        _errorDetail = errorDetail {
-    _validateErrorCode(_errorCode);
-    if (_errorDetail is! String) {
-      throw new ArgumentError.value(
-          _errorDetail, "errorDetail", "Must be a String");
-    }
+      : result = null,
+        errorCode = errorCode,
+        errorDetail = errorDetail {
+    _validateErrorCode(errorCode);
+    // TODO: When NNBD is complete, delete the following line.
+    checkNotNullable(errorDetail, "errorDetail");
   }
 
-  /// Invalid method parameter(s) error code.
-  @deprecated
-  static const kInvalidParams = invalidParams;
-
-  /// Generic extension error code.
-  @deprecated
-  static const kExtensionError = extensionError;
-
-  /// Maximum extension provided error code.
-  @deprecated
-  static const kExtensionErrorMax = extensionErrorMax;
-
-  /// Minimum extension provided error code.
-  @deprecated
-  static const kExtensionErrorMin = extensionErrorMin;
-
   /// Invalid method parameter(s) error code.
   static const invalidParams = -32602;
 
@@ -80,34 +66,26 @@ class ServiceExtensionResponse {
   }
 
   static _validateErrorCode(int errorCode) {
-    if (errorCode is! int) {
-      throw new ArgumentError.value(errorCode, "errorCode", "Must be an int");
-    }
-    if (errorCode == invalidParams) {
-      return;
-    }
+    // TODO: When NNBD is complete, delete the following line.
+    checkNotNullable(errorCode, "errorCode");
+    if (errorCode == invalidParams) return;
     if ((errorCode >= extensionErrorMin) && (errorCode <= extensionErrorMax)) {
       return;
     }
     throw new ArgumentError.value(errorCode, "errorCode", "Out of range");
   }
 
-  // ignore: unused_element, called from runtime/lib/developer.dart
-  bool _isError() => (_errorCode != null) && (_errorDetail != null);
+  /// Determines if this response represents an error.
+  bool isError() => (errorCode != null) && (errorDetail != null);
 
   // ignore: unused_element, called from runtime/lib/developer.dart
   String _toString() {
-    if (_result != null) {
-      return _result;
-    } else {
-      assert(_errorCode != null);
-      assert(_errorDetail != null);
-      return json.encode({
-        'code': _errorCode,
-        'message': _errorCodeMessage(_errorCode),
-        'data': {'details': _errorDetail}
-      });
-    }
+    return result ??
+        json.encode({
+          'code': errorCode!,
+          'message': _errorCodeMessage(errorCode!),
+          'data': {'details': errorDetail!}
+        });
   }
 }
 
@@ -134,32 +112,79 @@ typedef Future<ServiceExtensionResponse> ServiceExtensionHandler(
 /// Because service extensions are isolate specific, clients using extensions
 /// must always include an 'isolateId' parameter with each RPC.
 void registerExtension(String method, ServiceExtensionHandler handler) {
-  if (method is! String) {
-    throw new ArgumentError.value(method, 'method', 'Must be a String');
-  }
+  // TODO: When NNBD is complete, delete the following line.
+  checkNotNullable(method, 'method');
   if (!method.startsWith('ext.')) {
     throw new ArgumentError.value(method, 'method', 'Must begin with ext.');
   }
   if (_lookupExtension(method) != null) {
     throw new ArgumentError('Extension already registered: $method');
   }
-  if (handler is! ServiceExtensionHandler) {
-    throw new ArgumentError.value(
-        handler, 'handler', 'Must be a ServiceExtensionHandler');
-  }
-  _registerExtension(method, handler);
+  // TODO: When NNBD is complete, delete the following line.
+  checkNotNullable(handler, 'handler');
+  final zoneHandler = Zone.current.bindBinaryCallback(handler);
+  _registerExtension(method, zoneHandler);
 }
 
-/// Post an event of [eventKind] with payload of [eventData] to the `Extension`
+/// Whether the "Extension" stream currently has at least one listener.
+///
+/// A client of the VM service can register as a listener
+/// on the extension stream using `listenStream` method.
+/// The extension stream has a listener while at least one such
+/// client has registered as a listener, and has not yet disconnected
+/// again.
+///
+/// Calling [postEvent] while the stream has listeners will attempt to
+/// deliver that event to all current listeners,
+/// although a listener can disconnect before the event is delivered.
+/// Calling [postEvent] when the stream has no listener means that
+/// no-one will receive the event, and the call is effectively a no-op.
+@pragma("vm:recognized", "other")
+@pragma("vm:prefer-inline")
+@pragma("vm:idempotent")
+external bool get extensionStreamHasListener;
+
+/// Post an event of [eventKind] with payload of [eventData] to the "Extension"
 /// event stream.
-void postEvent(String eventKind, Map eventData) {
-  if (eventKind is! String) {
-    throw new ArgumentError.value(eventKind, 'eventKind', 'Must be a String');
+///
+/// If [extensionStreamHasListener] is false, this method is a no-op.
+/// Override [stream] to set the destination stream that the event should be
+/// posted to. The [stream] may not start with an underscore or be a core VM
+/// Service stream.
+void postEvent(String eventKind, Map eventData, {String stream = 'Extension'}) {
+  const destinationStreamKey = '__destinationStream';
+  // Keep protected streams in sync with `streams_` in runtime/vm/service.cc
+  // `Extension` is the only stream that should not be protected here.
+  final protectedStreams = <String>[
+    'VM',
+    'Isolate',
+    'Debug',
+    'GC',
+    '_Echo',
+    'HeapSnapshot',
+    'Logging',
+    'Timeline',
+    'Profiler',
+  ];
+
+  if (protectedStreams.contains(stream)) {
+    throw ArgumentError.value(
+        stream, 'stream', 'Cannot be a protected stream.');
+  } else if (stream.startsWith('_')) {
+    throw ArgumentError.value(
+        stream, 'stream', 'Cannot start with an underscore.');
   }
-  if (eventData is! Map) {
-    throw new ArgumentError.value(eventData, 'eventData', 'Must be a Map');
+
+  if (!extensionStreamHasListener) {
+    return;
   }
-  String eventDataAsString = json.encode(eventData);
+  // TODO: When NNBD is complete, delete the following two lines.
+  checkNotNullable(eventKind, 'eventKind');
+  checkNotNullable(eventData, 'eventData');
+  checkNotNullable(stream, 'stream');
+  Map mutableEventData = Map.from(eventData); // Shallow copy.
+  mutableEventData[destinationStreamKey] = stream;
+  String eventDataAsString = json.encode(mutableEventData);
   _postEvent(eventKind, eventDataAsString);
 }
 
@@ -170,5 +195,5 @@ external void _postEvent(String eventKind, String eventData);
 // these into Dart code unless you can ensure that the operations will can be
 // done atomically. Native code lives in vm/isolate.cc-
 // LookupServiceExtensionHandler and RegisterServiceExtensionHandler.
-external ServiceExtensionHandler _lookupExtension(String method);
+external ServiceExtensionHandler? _lookupExtension(String method);
 external _registerExtension(String method, ServiceExtensionHandler handler);
